



#include "main_modbus.h"



int nb_trames_valides = 0;
int nb_trames_invalides = 0;
int nb_trames_non_gerees = 0;



// Décode la trame reçue
void Decode_Trame_Modbus(unsigned char *tab, int len)
{
	static int num_esclave = -1, requete = 1; // requete vaut 0 ou 1 et alterne à chaque trame, permet de savoir si la trame est une requête du maître ou une réponse de l'esclave
	static int registre_demande = -1; // l'adresse du registre demandé dans une requête n'est pas inclus dans certaines réponses, c'est pourquoi il faut le mémoriser
	
	// Le numéro de l'esclave a changé
	if(tab[0] != num_esclave && Trame_Valide(tab, len))
	{
		num_esclave = tab[0];
		#if AFFICHE_TRAMES || AFFICHE_REGISTRES
		Aff_Nouvel_Esclave(num_esclave);
		#endif
			
		requete = 1; // si on change d'esclave, on recalle l'alternance requête/réponse sur requête : c'est forcément le maître qui parle
	}
	
	
	
	
    // On vérifie la cohérence entre le flag d'alternance requête/réponse et la valeur devinée suite au décodage de la trame
	/*if(requete == Is_Request(tab, len))
	{
		COULEUR(VERT);
		printf("- ");
	}
	else
	{
		COULEUR(ROUGE);
		printf("? ");
	}
	COULEUR(BLANC);*/
	
	
	
	
	// Affichage de la trame
	#if AFFICHE_TRAMES
	Aff_Trame(tab, len); // Affiche la trame sur une seule ligne
	#endif
	
	
	// Vérification validité
	if(!Trame_Valide(tab, len))
	{
		#if AFFICHE_TRAMES
		Aff_Trame_Invalide();
		#endif
		
		nb_trames_invalides++;
		registre_demande = -1;
		return;
	}
	nb_trames_valides++;
	
	
	
	// READ HOLDING REGISTER
	if(tab[1] == 0x03)
	{
		if(requete)//Is_Request(tab, len)) // Demande du maître
		{
			#if AFFICHE_TRAMES
			COULEUR(MAGENTA);
			printf("--> Demande de %d registre(s) à partir de l'adresse 0x%x à l'esclave %d\n", (tab[4]<<8) | tab[5], (tab[2]<<8) | tab[3] , tab[0]);
			#endif
			registre_demande = (tab[2]<<8) | tab[3];
		}
		else if(registre_demande != -1) // Réponse de l'esclave ET on a bien pu mémoriser le registre demandé
		{
			#if AFFICHE_TRAMES
			COULEUR(VERT);
			printf(" <- Réponse de %d octet(s) (%d registre(s)) de l'esclave %d qui donne la valeur de(s) registre(s) à partir de 0x%x\n", tab[2], tab[2]/2, tab[0], registre_demande);
			#endif
			
			Maj_Bloc_Registres(tab[0], registre_demande, &tab[3], tab[2]/2, REGISTRE_LU); // numero d'esclave, adresse modbus du premier registre, pointeur vers les registres, nombre de registres
		}
	}
	
	
	// WRITE MULTIPLE REGISTERS
	else if(tab[1] == 0x10)
	{
		if(requete)//Is_Request(tab, len)) // Demande du maître
		{
			#if AFFICHE_TRAMES
			COULEUR(MAGENTA);
			printf("--> Ecriture de %d registre(s) à partir de l'adresse 0x%x à l'esclave %d\n", (tab[4]<<8) | tab[5], (tab[2]<<8) | tab[3] , tab[0]);
			#endif
			
			Maj_Bloc_Registres(tab[0], (tab[2]<<8) | tab[3], &tab[7], (tab[4]<<8) | tab[5], REGISTRE_ECRIT); // numero d'esclave, adresse modbus du premier registre, pointeur vers les registres, nombre de registres, type d'accès (lecture ou écriture)
		}
		else // Réponse de l'esclave
		{
			#if AFFICHE_TRAMES
			COULEUR(VERT);
			printf(" <- Réponse suite à l'écriture de %d registre(s) de l'esclave %d\n", (tab[4]<<8)|tab[5], tab[0]);
			#endif
		}
		//printf("\n");
	}
	
	
	// AUTRE
	else
	{
		#if AFFICHE_TRAMES
		COULEUR(ROUGE);
		printf("Type de trame non géré !!\n");
		#endif
		nb_trames_non_gerees++;
	}
	
	
	
    #if AFFICHE_TRAMES
	COULEUR(BLANC);
	#endif
	requete = !requete;
}








// Enregistre dans la liste des registres ceux lus ou écrits par la trame
void Maj_Bloc_Registres(uint8_t num_esclave, uint16_t adresse_premier_registre, uint8_t *data, uint16_t nb_registres, int type_acces)
{
    uint16_t i;
		
	//printf("  --> Récupération de %d registres\n", nb_registres);
	
    for(i = 0; i < nb_registres; i++)
	{
		uint16_t nouvelle_valeur = (data[i*2]<<8) | data[i*2+1];
		
		Recoit_Registre(num_esclave, adresse_premier_registre + i, nouvelle_valeur, type_acces);		
	}
}




// Renvoie 1 si la trame est valide (CRC OK), 0 sinon
int Trame_Valide(unsigned char *tab, int len)
{
  uint16_t crc_calcule, crc_trame;
	if(len < 4)
	    return 0; // taille mini 4 octets
	
	crc_calcule = crc16(tab, len-2); // On calcule le CRC de la trame à laquelle on a bien sûr enlevé les deux derniers octets de CRC
	crc_trame = (tab[len-1] << 8) | tab[len-2];
	if(crc_calcule == crc_trame)
	    return 1;
	else
	{
		//printf("CRC trame = 0x%x, CRC calculé = 0x%x\n", crc_trame, crc_calcule);
	  return 0;
	}
}








// Renvoie le CRC Modbus de la trame passée en paramètre
uint16_t crc16(uint8_t *data, uint8_t len)
{
    int i, j;
	uint16_t poly = 0xA001;
	uint16_t crc = 0xFFFF;
	
	for(i = 0; i < len; i++)
	{
		crc ^= (0x00FF & data[i]);
		for(j = 0; j < 8; j++)
		{
			if (crc & 0x0001)
				crc = ((crc >> 1) & 0xFFFF) ^ poly;
			else
				crc = ((crc >> 1) & 0xFFFF);
		}
    }
	return crc;
}




// Renvoie 1 si la trame passée en paramètre est une demande du maître
// Renvoie 0 si c'est une réponse de l'esclave
// Renvoie -1 si la trame est invalide

// EN PRATIQUE CE N'EST PAS POSSIBLE DE DETERMINER CELA DE MANIERE CERTAINE. DONC ON COMMENTE LA FONCTION ET ON NE S'EN SERT PLUS
/*int Is_Request(unsigned char *tab, int len)
{
    if(!Trame_Valide(tab, len))
	    return -1;
	
	// SI on est ici c'est que la trame est valide, donc len >= 4. On peut lire tab[2] sans risque

	if(tab[1] == 0x03) // READ HOLDING REGISTERS
	{
		// On regarde si le 3è caractère est la longueur de la trame, diminuée de l'adresse de l'esclave, du type d'action et des 2 octets de checksum, auquel cas c'est une réponse
		// il faut par contre que les octets 4 et 5 ne contiennent pas 
		if(len == tab[2] + 3 + 2)
			return 0; // réponse
		else
			return 1; // requête
	}
	else if(tab[1] == 0x10) // WRITE MULTIPLE REGISTER
	{
	    if(len == 8) // c'est une réponse
		    return 0;
		else
		    return 1;
	}
	
	return 0;
}*/

