
#include "main_modbus.h"



// Gros tableau utilisé pour stocker tous les registres de tous les esclaves
struct registre_t liste_registres[NB_MAX_ESCLAVES][NB_MAX_ADRESSES]; // premier index : numéro de l'esclave, second index : numéro du registre




// Réinitialise l'affichage
void Aff_Init()
{
	uint64_t ts1 = tsus(), ts2;
	
	printf("\033[2J");
	Nettoie_Registres();
	Aff_Registres();
	ts2 = tsus();
	printf("\n\n>>> Affichage en %ldus (doit être < 1500us !!) <<<\n", ts2-ts1);
}







// Affiche la trame passée en paramètre sur une seule ligne, sans retour chariot final
void Aff_Trame(unsigned char * tab, int len)
{
	int i;
	
	//printf("[ ");
	for(i = 0; i < len-2; i++) // Checksum non affiché !
		printf("0x%02x ", tab[i]);
		//printf("%d ", tab[i]);
	
	//printf(" ] ");
	
}




// Affiche le message en cas de trame invalide
void Aff_Trame_Invalide()
{
	COULEUR(ROUGE);
	printf("Trame invalide !\n");
}





// Initialise tous les registres à l'état vide : on n'a pas encore lu ou écrit leur valeur, on ne sait pas c equ'ils contiennent
void Init_Registres()
{
   uint16_t i, j;
	for(i = 0; i < NB_MAX_ESCLAVES; i++)
	{
	    for(j = 0; j < NB_MAX_ADRESSES; j++)
		{
			liste_registres[i][j].valeur = 0;
			liste_registres[i][j].type_acces  = AUCUN_ACCES;
			liste_registres[i][j].dernier_rx = 0;
		}
	
	}
}



// Prépare l'affichage pour un nouvel esclave
void Aff_Nouvel_Esclave(int nouvel_esclave)
{
	#if AFFICHE_TRAMES
	printf("\n"); // Affiche une nouvelle ligne, juste pour la mise en forme
	#endif
	
	#if AFFICHE_REGISTRES
	if(nouvel_esclave == 1)
		Aff_Init(); // Réinitialise l'affichage complet
	#endif
	
}



// Supprime de la liste les registres trop anciens
void Nettoie_Registres()
{
	uint16_t i, j;

    // On parcourt tous les registres de chaque esclave
	for(i = 0; i < NB_MAX_ESCLAVES; i++)
	{
	    for(j = 0; j < NB_MAX_ADRESSES; j++)
		{
		    if(time(NULL) > liste_registres[i][j].dernier_rx + TIMEOUT_REGISTRE) // On connait la valeur de ce registre
			{
			    liste_registres[i][j].valeur = 0;
			    liste_registres[i][j].dernier_rx = 0;
			    liste_registres[i][j].type_acces = AUCUN_ACCES;
			}
		}
	}
}



// Reçoit la valeur d'un registre et la stocke dans le tableau
void Recoit_Registre(int num_esclave, int adresse, int valeur, int type_acces)
{
	if(adresse < NB_MAX_ADRESSES && num_esclave < NB_MAX_ESCLAVES)
	{
		if(liste_registres[num_esclave][adresse].valeur != valeur || liste_registres[num_esclave][adresse].dernier_rx == 0)
			liste_registres[num_esclave][adresse].dernier_cht = time(NULL);
		
		liste_registres[num_esclave][adresse].dernier_rx = time(NULL);
		liste_registres[num_esclave][adresse].valeur = valeur;
		liste_registres[num_esclave][adresse].type_acces = type_acces;
	}
	
	#if AFFICHE_TRAMES
	printf("  %d Reg 0x%x = 0x%x\n", num_esclave, adresse, liste_registres[num_esclave][adresse].valeur);
	#endif 
}



// Affiche la liste des registres dont la valeur est connue, tous esclaves confondus
void Aff_Registres()
{
	#if AFFICHE_REGISTRES
    uint16_t i, j, index_registre;
    int16_t liste_registres_par_esclave[NB_MAX_ESCLAVES][NB_MAX_ADRESSES]; // Une valeur >= 0 indique un registre qui est connu, la première case à -1 donne la fin de la liste
    int16_t liste_registres_total[NB_MAX_ADRESSES] = {-1}; // Une valeur >= 0 indique un registre qui est connu, la première case à -1 donne la fin de la liste
	
    // On regarde quels sont les registres dont un connait la valeur pour chaque esclave
	for(i = 0; i < NB_MAX_ESCLAVES; i++)
	{
	    index_registre = 0;
	    for(j = 0; j < NB_MAX_ADRESSES; j++)
		{
		    if(liste_registres[i][j].dernier_rx > 0) // On connait la valeur de ce registre
			{
			    liste_registres_par_esclave[i][index_registre] = j;
				index_registre++;
			}
			liste_registres_par_esclave[i][index_registre] = -1;
		}
	}
	
	
	
	/*for(i = 0; i < NB_MAX_ESCLAVES; i++)
	{
		printf("Esclave %d, registres connus : ", i);
		for(j = 0; j < NB_MAX_ADRESSES && liste_registres_par_esclave[i][j] != -1; j++)
		{
			printf("0x%x  ", liste_registres_par_esclave[i][j]);
		}
		printf("\n");
	}*/

	
	// A ce stade, on a donc un tableau 2D 'liste_registres_par_esclave' qui contient, pour chaque esclave, la liste des registres qu'il connaît. 	
	
	// Maintenant, on fusionne la liste de tous les registres de chaque esclave pour savoir quoi afficher
	//uint16_t tab_indices[NB_MAX_ESCLAVES] = {0}; // Tableau qui contient l'index courant de chaque esclave
	index_registre = 0;
	int32_t derniere_valeur_ecrite = -1;
	while(1)
	{
	    // On regarde quelle est la valeur minimale parmi tous les esclaves mais quand même supérieure à la dernier valeur écrite dans la liste fusionnée
		uint16_t valeur_min = 65535;
		for(i = 0; i < NB_MAX_ESCLAVES; i++)
		{
		    for(j = 0; j < NB_MAX_ADRESSES; j++)
			{
				if(liste_registres_par_esclave[i][j] <0) // On est arrivé au bout pour cet esclave
				    break;
				
				// On regarde la valeur mini pour cet esclave
				if(liste_registres_par_esclave[i][j] <= valeur_min && liste_registres_par_esclave[i][j] > derniere_valeur_ecrite)
					valeur_min = liste_registres_par_esclave[i][j];
			}
		}
		if(valeur_min == 65535)
		    break; // on a terminé la fusion
		
		liste_registres_total[index_registre] = valeur_min;
		derniere_valeur_ecrite = valeur_min;
		index_registre++;
	}
	liste_registres_total[index_registre] = -1;
	
	// on a maintenant le tableau 'liste_registres_total' qui contient la liste de tous les registres de tous les esclaves confondus. La fin de la liste est indiquée par la valeur -1
	
	
	
	// Affichage à proprement parler !
	
	// On affiche les numéros des esclaves
	for(i = 1; i < NB_MAX_ESCLAVES; i++) // On parcourt les esclaves, à partir du n°1
	{
		printXY(i * 20, 0);
		printf("Esclave %d", i);
	}
	
	// On parcourt chaque registre existant
	for(i = 0; liste_registres_total[i] != -1; i++)
	{
		// On affiche les légendes des registres
		COULEUR(BLANC);
		printXY(0, 2+i);
		printf("Reg %d", liste_registres_total[i]);
			
		// On affiche les valeurs de ses registres
		for(j = 0; j < NB_MAX_ESCLAVES; j++) // On parcourt les esclaves
		{			
				struct registre_t structure_registre = liste_registres[j][liste_registres_total[i]]; // On récupère la structure de ce registre
				
				//if(i == 0 && j == 1)
				 //   printf("Esclave %d, registre %d : 0x%x. Direct ds tableau : 0x%x\n", j, liste_registres_total[i], structure_registre.valeur, (uint16_t)(liste_registres[1][0].valeur));
				
				if(structure_registre.type_acces == REGISTRE_LU)
				{
					printXY(j * 20, 2+i);
					COULEUR(VERT);
					printf("R ");
					
					if(time(NULL) > structure_registre.dernier_cht + DUREE_NOUVEAU_REGISTRE)
						COULEUR(BLANC);
					else
						COULEUR(MAGENTA);
					
					printf("0x%x", structure_registre.valeur);
				}
				else if(structure_registre.type_acces == REGISTRE_ECRIT)
				{
					printXY(j * 20, 2+i);
					COULEUR(ROUGE);
					printf("W ");
					
					if(time(NULL) > structure_registre.dernier_cht + DUREE_NOUVEAU_REGISTRE)
						COULEUR(BLANC);
					else
						COULEUR(MAGENTA);
					printf("0x%x", structure_registre.valeur);
				}
		}
	}
	
	fflush(stdout);
	#endif
}







